#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>

#include <sndfile.h>

#include <libharkio3.h>
#define BUF_LEN 2048


// for debug out
void help(void)
{
    printf("                                                                    \n");
    printf("                                                                    \n");
    printf("[NAME:]                                                             \n");
    printf("  harktoolcli-gentsp                                                \n");
    printf("                                                                    \n");
    printf("[Description:]                                                      \n");
    printf("  TSP signal generating tool. The file is saved as 32bit float wave.\n");
    printf("                                                                    \n");
    printf("[Usage :]                                                           \n");
    printf(" harktoolcli-gentsp output_filename [options]                       \n");
    printf("                                                                    \n");
    printf("[options]                                                           \n");
    printf("      -n: Number of TSPs                    (-n 16)    [integer]    \n");
    printf("      -o: Offset of the TSP                 (-o 0)     [integer]    \n");
    printf("      -r: Sampling rate of TSP              (-p 16000) [integer]    \n");
    printf("      -s: input tsp raw (32 bit float) file            [string]     \n");
    printf("                                                                    \n");
    printf("                                                                    \n");
}








/*------------------------------------------------------*/
// read 16384.little_endian.tsp
/*------------------------------------------------------*/
float* readRawTSP(char *path, long *tsplen)
{
    FILE* fp = fopen(path, "rb");
    long size, retsize;
    char buf[BUF_LEN];
    float *tsp;

    if (fp == NULL) {
        harkio_Log_FileOpenFailed(path);
        return NULL;
    }

    // check file size
    fseek(fp, 0, SEEK_END);
    size = ftell(fp) / sizeof(float);
    rewind(fp);

    snprintf(buf, BUF_LEN, "TSP file includes %ld samples.", size);
    harkio_Log_printf(harkio_Log_Debug, buf);

    // read files
    tsp = (float *) malloc(size * sizeof(float));
    retsize = fread(tsp, sizeof(float), size, fp);

    fclose(fp);
    fp = NULL;

    if (retsize != size) {
        harkio_Log_FileOpenFailed(path);
        free(tsp);
        return NULL;
    }

    *tsplen = size;
    return tsp;
}




/*------------------------------------------------------*/
// save float as 32bit float wav
/*------------------------------------------------------*/
int saveas(float *tsp, int tsplen, int samplingrate, char *fname)
{
    SF_INFO sinfo;
    SNDFILE *sf;

    // setup header
    sinfo.frames = tsplen;
    sinfo.channels = 1;
    sinfo.samplerate = samplingrate;
    sinfo.format = SF_FORMAT_WAV|SF_FORMAT_FLOAT;

    // format check
    if (sf_format_check(&sinfo) == 0) {
        harkio_Log_printf(harkio_Log_Error, "Format Error.");
    }

    // write data
    sf = sf_open(fname, SFM_WRITE, &sinfo);
    sf_writef_float(sf, tsp, tsplen);
    sf_close(sf);
    return EXIT_SUCCESS;
}






/*------------------------------------------------------*/
//
/*------------------------------------------------------*/
int main(int argc, char *argv[])
{
    int option;

    // arguments
    int numOfTSP = 16;
    int offset = 0;
    int samplingrate = 16000;
    char *tsppath = "16384.little_endian.tsp";
    int isTSPPATHalloced = 0;
    char outfile[BUF_LEN];
    char buf[BUF_LEN];

    // tsp buffers
    float *origtsp, *gentsp;
    long tsplen;

    int i;

    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help();
        exit(EXIT_FAILURE);
    }

    // help
    if (strcmp(argv[1] , "--help") == 0) {
        help();
        exit(EXIT_SUCCESS);
    }

    // store output file name
    strncpy(outfile, argv[1], BUF_LEN);

    // parse option
    while ((option = getopt(argc, argv, "n:o:r:s:")) != -1) {
        switch (option) {
        case 'n':
            numOfTSP = atoi(optarg);
            break;

        case 'o':
            offset = atoi(optarg);
            break;

        case 'r':
            samplingrate = atoi(optarg);
            break;

        case 's':
            tsppath = (char *)malloc(strlen(optarg) + 1);
            memcpy(tsppath, optarg, strlen(optarg));
            isTSPPATHalloced = 1;
            break;

        case '?':
            help();
            exit(EXIT_FAILURE);
        }
    }

    // debug out
    snprintf(buf, BUF_LEN, "n: %d", numOfTSP);
    harkio_Log_printf(harkio_Log_Debug, buf);
    snprintf(buf, BUF_LEN, "o: %d", offset);
    harkio_Log_printf(harkio_Log_Debug, buf);
    snprintf(buf, BUF_LEN, "r: %d", samplingrate);
    harkio_Log_printf(harkio_Log_Debug, buf);
    snprintf(buf, BUF_LEN, "s: %s", tsppath);
    harkio_Log_printf(harkio_Log_Debug, buf);

    // read original TSP
    if ((origtsp = readRawTSP(tsppath, &tsplen)) == NULL) {
        harkio_Log_printf(harkio_Log_Error, "Reading raw tsp failed.");
        free(origtsp);
        exit(EXIT_FAILURE);
    }

    // allocate memory for multiple TSP + offset
    gentsp = (float *) malloc(sizeof(float) * (tsplen * numOfTSP + offset));

    // add offset
    memset(gentsp, 0, sizeof(float) * offset);

    // copy tsps
    for (i=0; i<numOfTSP; i++) {
        memcpy(gentsp + offset + tsplen * i,
               origtsp, sizeof(float) * tsplen);
    }

    // save file
    if (saveas(gentsp, tsplen * numOfTSP + offset,
               samplingrate, outfile) == EXIT_FAILURE) {
        harkio_Log_printf(harkio_Log_Error, "Save file failed");
    }

    // free them
    free(origtsp);
    free(gentsp);

    if (isTSPPATHalloced)
        free(tsppath);

    harkio_Log_printf(harkio_Log_Debug, "File generated.");

    exit(EXIT_SUCCESS);

}
